//	Exstrata, javascript api, Version 1.00, 2015-01-14, Fertile Media

function Exstrata (applicationKey) {
	if (applicationKey) {
		this.applicationKey = applicationKey;
	} else {
		this.applicationKey = 'unknown_application';
	}
	this.timeDifference = 0;
	
	this._liveView = false; // If this is true, the api continuously updates itself with the live actions
	this.patterns = {};
	this.actions = [];
	
	this._liveControl = false;					// When this is true liveControl commands are allowed
	this._liveControlToken = null;				// A unique token that gives the right to a position in the queue and to liveControl when the time comes.
	this._liveControlQueuePosition = 0;			// Current position in liveControl queue. (0 = not in queue, 1+ = number of "people" in line before user)
	this._liveControlStartTime = 0; 			// Timestamp (milliseconds) when liveControl starts
	this._liveControlDuration = 0;				// Milliseconds indicating total length of liveControl time
	this._liveControlSecondsTillStart = 0;		// Seconds until liveControl starts
	this._liveControlSecondsTillEnd = 0;		// Seconds until liveControl ends
	
	this.livedataUpdateTimer = 0;
	this.liveControlUpdateTimer = 0;
	
	this.tickCounter = 0;
	
	this.ticker = setInterval(jQuery.proxy(this.tick, this), 100);
}

//	Services

//	Base method for calling a service
Exstrata.prototype.callService = function (url, data, callback) {
	data.applicationKey = this.applicationKey,
	jQuery.ajax({
		url: url,
		type: 'POST',
		data: data,
		dataType: "json"
	}).done(jQuery.proxy(function(response){
		if (response.result) {
			if (callback) {
				callback(response);
			}
		} else {
			alert(response.feedback);	
		}
		jQuery(this).trigger('serviceEnd');
	}, this)).fail(jQuery.proxy(function(response){
		console.log('Service Fail: '+ response.responseText);
	}, this));
}

Exstrata.prototype.fetchLivedata = function (actionIds, liveControlToken, callback) {
	var data = {
		'liveControlToken': liveControlToken,
		'actionIds': actionIds
	};
	this.livedataUpdateTimer = 0;
	this.callService('http://exstrata.nl/control/api/fetch_livedata.php', data, jQuery.proxy(function(response){
		this.setCurrentServerTime(response.serverTime);
		this.addPatterns(response.patterns);
		this.addActions(response.actions);
		if (response.queueUpdate) {
			this.processQueueUpdate(response);
		}
		if (callback) {
			callback(response);
		}
	}, this));
	return true;
}

Exstrata.prototype.fetchPublicPatterns = function (callback) {
	var data = {};
	this.callService('http://exstrata.nl/control/api/fetch_public_patterns.php', data, callback);
}

/*
	subscribeToLiveControl
	To be able to play patterns live you first need to subscribe to the queue
*/
Exstrata.prototype.subscribeToLiveControl = function (callback) {
	if (!this.liveControlToken()) {
		var data = {};
		this.callService('http://exstrata.nl/control/api/subscribe_to_live_control.php', data, jQuery.proxy(function(response){
			this.setCurrentServerTime(response.serverTime);
			this.processLiveControlUpdate(response);
			this.setLiveControlDuration(response.liveControlDuration);
			this.setLiveControlToken(response.liveControlToken);
		}, this));
		return true;
	} else {
		return false;
	}
}

/*
	unsubscribeFromLiveControl
	When leaving the queue, it's nice to let us know. ;)
*/
Exstrata.prototype.unsubscribeFromLiveControl = function (callback) {
	if (this.liveControlToken()) {
		var data = {
			'liveControlToken': this.liveControlToken()
		};
		this.callService('http://exstrata.nl/control/api/unsubscribe_from_live_control.php', data, jQuery.proxy(function(response){
			this.setLiveControlToken(false);
			if (callback) {
				callback(response.result);
			}
		}, this));
		return true;
	} else {
		return false;
	}
}

/*
	fetchLiveControlUpdate
	To update the liveControl information, and let the system know you're still here, use this service
	(api automatically keeps updating the liveControl data when subscribed to queueu)
*/
Exstrata.prototype.fetchLiveControlUpdate = function (callback) {
	//	Fetch liveControl updates only when required
	if (this.liveControlToken() && this.liveControlQueuePosition() > 0) {
		var data = {
			'liveControlToken': this.liveControlToken()
		};
		this.liveControlUpdateTimer = 0;
		this.callService('http://exstrata.nl/control/api/fetch_live_control_update.php', data, jQuery.proxy(function(response){
			this.processLiveControlUpdate(response);
			if (callback) {
				callback(response.result);
			}
		}, this));
		return true;
	} else {
		return false;
	}
}

/*
	playPatternById
	Send a pattern id to the server to play immediatly.
	Can only be used when live controlling, i.e. first in queue
	Returns false if not first in queue
*/
Exstrata.prototype.playPatternById = function (patternId, callback) {
	if (this.liveControl()) {
		var data = {
			'liveControlToken': this.liveControlToken(),
			'pattern_id': patternId
		};
		this.callService('http://exstrata.nl/control/api/play_pattern_by_id.php', data, jQuery.proxy(function(response){
			if (callback) {
				callback(response.result);
			}
		}), this);
	} else {
		return false;
	}
}

/*
	playPattern
	Send a pattern to the server to play immediatly.
	Can only be used when live controlling, i.e. first in queue
	Returns false if not first in queue
*/
Exstrata.prototype.playPattern = function (pattern, callback) {
	if (this.liveControl()) {
		var data = {
			'liveControlToken': this.liveControlToken(),
			'pattern': pattern
		};
		this.callService('http://exstrata.nl/control/api/play_pattern.php', data, jQuery.proxy(function(response){
			if (callback) {
				callback(response.result);
			}
		}), this);
		return true;
	} else {
		return false;
	}
}

//	currentServerTime

Exstrata.prototype.setCurrentServerTime = function (time) {
	var now = new Date();
	this.timeDifference = now.getTime() - time;
}

Exstrata.prototype.getCurrentServerTime = function () {
	var now = new Date();
	return now.getTime() - this.timeDifference;
}

//	Patterns (Library)

Exstrata.prototype.addPattern = function (pattern) {
	this.patterns[pattern.id] = pattern;
}

Exstrata.prototype.addPatterns = function (patterns) {
	for (i in patterns) {
		this.addPattern(patterns[i]);
	}
}

Exstrata.prototype.getPatternById = function (id) {
	return this.patterns[id];
}

//	Actions (library)

Exstrata.prototype.addAction = function (action) {
	
	for (var i = 0; i < this.actions.length; i++) {
		//	Override itself
		if (this.actions[i].id == action.id) {
			return;
		}
	}
	
	for (var i = 0; i < this.actions.length; i++) {
		//If the queue's start time is greater than the object start time it should be spliced on that spot.
		if (this.actions[i].start_time > action.start_time) {
			this.actions.splice(i, 0, action);
			return;
		}
	}
	
	this.actions.push(action);
}

Exstrata.prototype.addActions = function (actions) {
	for (i in actions) {
		this.addAction(actions[i]);
	}
}

Exstrata.prototype.getActionIds = function () {
	var ids = [];
	for (var i in this.actions) {
		ids.push(this.actions[i].id);
	}
	return ids;
}

Exstrata.prototype.cleanActions = function () {
	var doneCleaning = false;
	while (!doneCleaning) {
		if (this.actions[0] && parseFloat(this.actions[0].start_time) + parseFloat(this.actions[0].duration_ms) < this.getCurrentServerTime()) {
			this.actions.shift();
		} else {
			doneCleaning = true;
		}
	}	
}

//	activeAction

Exstrata.prototype.setActiveAction = function (action) {
	if (action != this.activeAction) {
		console.log('setActiveAction('+ Date() +')');
		console.log(action);
		this.activeAction = action;
		jQuery(this).trigger('activeActionChanged');
		
		//	Dispatch the activeAction along with the pattern
		var liveAction = Object();
		if (this.activeAction) {
			liveAction.start_time = this.activeAction.start_time;
			//liveAction.start_time = this.activeAction.start_time;
			
			var pattern = this.getPatternById(this.activeAction.cached_pattern_id);
			console.log(pattern);
			if (pattern) {
				liveAction.frames = pattern.frames;
			}
		}
		
		var event = jQuery.Event("liveActionChanged");
		event.liveAction = liveAction;
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.updateActiveAction = function () {
	//	Clean out any actions that will no longer be played
	this.cleanActions();
	
	for (var i = (this.actions.length -1); i >= 0; i--) {
		if (this.actions[i].start_time <= this.getCurrentServerTime()) {
			this.setActiveAction(this.actions[i]);
			return;
		}
	}
}

//	Tick

Exstrata.prototype.tick = function () {
	this.tickCounter++;
	
	if (this.liveView()) {
		
		//	Fetch livedata
		if (this.livedataUpdateTimer > 5) {
			this.updateLivedata();
		}
		this.livedataUpdateTimer++;
		
		//	Check for new action
		this.updateActiveAction();
		
	}
	
	if (this.liveControlToken()) {
		//	Fetch queue status (when in livestate, updateLivedata will do this)
		if (this.liveControlQueuePosition() != 0) {
			if (this.liveControlUpdateTimer > 10 && !this.liveView()) {
				this.fetchLiveControlUpdate();
			}
			this.liveControlUpdateTimer++;
		}
		
		//	Update the queue wait time every second
		if ((this.tickCounter % 5) === 0) {
			this.updateLiveControlSecondsTillStart();
			this.updateLiveControlSecondsTillEnd();
		}
	}
}

//	liveControl

Exstrata.prototype.setLiveControl = function (value) {
	if (value != this.liveControl()) {
		this._liveControl = value;
		
		var event = jQuery.Event("liveControlChange");
		event.liveControl = this.liveControl();
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.getLiveControl = function () {
	return this._liveControl;
}

Exstrata.prototype.liveControl = function () {
	return this.getLiveControl();
}

Exstrata.prototype.updateLiveControl = function () {
	if (this.liveControlToken()) {
		if (this.liveControlQueuePosition() == 0) {
			if (this.liveControlStartTime() != 0) {
				if (this.liveControlDuration() != 0) {
					if (this.liveControlSecondsTillStart() == 0) {
						if (this.liveControlSecondsTillEnd() > 0) {
							this.setLiveControl(true);
							return;
						}
					}
				}
			}
		}
	}
	this.setLiveControl(false);
}

//	liveControlToken

Exstrata.prototype.setLiveControlToken = function (value) {
	if (value != this.liveControlToken) {
		this._liveControlToken = value;
		
		this.processLiveControlToken();
		
		var event = jQuery.Event("liveControlTokenChange");
		event.liveControlToken = this.liveControlToken();
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.getLiveControlToken = function () {
	return this._liveControlToken;
}

Exstrata.prototype.liveControlToken = function () {
	return this.getLiveControlToken();
}

Exstrata.prototype.processLiveControlToken = function () {
	this.updateLiveControl();
	if (!this.getLiveControlToken()) {
		this.setLiveControlQueuePosition(0);
		this.setLiveControlStartTime(0);
		this.setLiveControlDuration(0);
	}
}

//	liveControlQueuePosition

Exstrata.prototype.setLiveControlQueuePosition = function (value) {
	if (value != this.liveControlQueuePosition()) {
		this._liveControlQueuePosition = value;
		
		this.processLiveControlQueuePosition();
		
		var event = jQuery.Event("liveControlQueuePositionChange");
		event.liveControlQueuePosition = this.liveControlQueuePosition();
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.getLiveControlQueuePosition = function () {
	return this._liveControlQueuePosition;
}

Exstrata.prototype.liveControlQueuePosition = function () {
	return this.getLiveControlQueuePosition();
}

Exstrata.prototype.processLiveControlQueuePosition = function () {
	this.updateLiveControl();
}

//	liveControlStartTime

Exstrata.prototype.setLiveControlStartTime = function (value) {
	if (value != this.liveControlStartTime()) {
		this._liveControlStartTime = value;
		
		this.processLiveControlStartTime();
	}
}

Exstrata.prototype.getLiveControlStartTime = function () {
	return this._liveControlStartTime;
}

Exstrata.prototype.liveControlStartTime = function () {
	return this.getLiveControlStartTime();
}

Exstrata.prototype.processLiveControlStartTime = function () {
	this.updateLiveControl();
	this.updateLiveControlSecondsTillStart();
	this.updateLiveControlSecondsTillEnd();
}

//	liveControlDuration

Exstrata.prototype.setLiveControlDuration = function (value) {
	if (value != this.liveControlDuration()) {
		this._liveControlDuration = value;
		
		this.processLiveControlDuration();
	}
}

Exstrata.prototype.getLiveControlDuration = function () {
	return this._liveControlDuration;
}

Exstrata.prototype.liveControlDuration = function () {
	return this.getLiveControlDuration();
}

Exstrata.prototype.processLiveControlDuration = function () {
	this.updateLiveControl();
	this.updateLiveControlSecondsTillEnd();
}

//	liveControlSecondsTillStart

Exstrata.prototype.setLiveControlSecondsTillStart = function (value) {
	if (value != this.liveControlSecondsTillStart()) {
		this._liveControlSecondsTillStart = value;
		
		this.processLiveControlSecondsTillStart();
		
		var event = jQuery.Event("liveControlSecondsTillStartChange");
		event.liveControlSecondsTillStart = this.liveControlSecondsTillStart();
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.getLiveControlSecondsTillStart = function () {
	return this._liveControlSecondsTillStart;
}

Exstrata.prototype.liveControlSecondsTillStart = function () {
	return this.getLiveControlSecondsTillStart();
}

Exstrata.prototype.processLiveControlSecondsTillStart = function () {
	this.updateLiveControl();
}

Exstrata.prototype.updateLiveControlSecondsTillStart = function () {
	var startTime = this.liveControlStartTime();
	if (startTime != 0) {
		var milliSecondsTillStart = startTime - this.getCurrentServerTime();
		var secondsTillStart = Math.ceil(milliSecondsTillStart / 1000);
		secondsTillStart = (secondsTillStart > 0) ? secondsTillStart : 0;
		this.setLiveControlSecondsTillStart(secondsTillStart);
	} else {
		this.setLiveControlSecondsTillStart(0);
	}
}


//	liveControlSecondsTillEnd (number of seconds before user live time ends

Exstrata.prototype.setLiveControlSecondsTillEnd = function (value) {
	if (value != this.liveControlSecondsTillEnd()) {
		this._liveControlSecondsTillEnd = value;
		
		this.processLiveControlSecondsTillEnd();
		
		var event = jQuery.Event("liveControlSecondsTillEndChange");
		event.liveControlSecondsTillEnd = this.liveControlSecondsTillEnd();
		jQuery(this).trigger(event);
	}
}

Exstrata.prototype.getLiveControlSecondsTillEnd = function () {
	return this._liveControlSecondsTillEnd;
}

Exstrata.prototype.liveControlSecondsTillEnd = function () {
	return this.getLiveControlSecondsTillEnd();
}

Exstrata.prototype.processLiveControlSecondsTillEnd = function () {
	if (this.liveControlToken && this.liveControlSecondsTillEnd == 0) {
		this.setLiveControlToken(null);
	}
	
	this.updateLiveControl();
}

Exstrata.prototype.updateLiveControlSecondsTillEnd = function () {
	var startTime = this.getLiveControlStartTime();
	if (startTime != 0) {
		var milliSecondsTillEnd = this.getLiveControlDuration() - (this.getCurrentServerTime() - startTime);
		var secondsTillEnd = Math.ceil(milliSecondsTillEnd / 1000);
		secondsTillEnd = (secondsTillEnd > 0) ? secondsTillEnd : 0;
		this.setLiveControlSecondsTillEnd(secondsTillEnd);
	} else {
		this.setLiveControlSecondsTillEnd(0);
	}
}

//	liveView

Exstrata.prototype.setLiveView = function (value) {
	this._liveView = value;
	this.updateLivedata();
}

Exstrata.prototype.getLiveView = function () {
	return this._liveView;
}

Exstrata.prototype.liveView = function () {
	return this.getLiveView();
}


//	processLiveControlUpdate

Exstrata.prototype.processLiveControlUpdate = function (serviceResponse) {
	this.setLiveControlQueuePosition(serviceResponse.liveControlQueuePosition);
	this.setLiveControlStartTime(serviceResponse.liveControlStartTime);
}

//


Exstrata.prototype.bind = function (eventname, callback) {
	return jQuery(this).bind(eventname, callback);
}

//	Public methods

Exstrata.prototype.updateLivedata = function () {
	
	//	Fetch queue update only once every 10 ticks (1 second) by  supplying queueToken
	var liveControlToken = '';
	if (this.updateQueueTimer > 10) {
		liveControlToken = this.getLiveControlToken();
	}
	
	this.fetchLivedata(this.getActionIds(), liveControlToken, jQuery.proxy(function(result){
		jQuery(this).trigger('updateLivedataDone');
	}, this));
}


Exstrata.prototype.runTestPattern = function () {
	this.bind('liveControlChange', jQuery.proxy(function(){
		if (this.liveControl()) {
			var pattern = this.getRandomPattern();
			this.playPattern(pattern, jQuery.proxy(function(){
			}), this);
			alert('Your testPattern has been sent!');
		}
	}), this);
	exstrata.bind('liveControlQueuePositionChange', function(event){
		alert('Your testPattern is currently in queue. Nr in queue: '+ event.liveControlQueuePosition);
	});
	
	alert('Starting test..')
	this.subscribeToLiveControl();
}

Exstrata.prototype.playRandomPattern = function () {
	if (this.getLiveControl()) {
		var pattern = this.getRandomPattern();
		this.playPattern(pattern);
	}
}

Exstrata.prototype.getRandomPattern = function () {
	var frames = 9 + Math.ceil(Math.random() * 90);
	var frames = 10;
	var zones = 80;
	//	Create and send a subpattern for each 10 frames (13 seems to be the maximum amount of any sent pattern)
	var subpatterns = 1;
	for (var i = 0; i < subpatterns; i++) {
	
		var subframes = Math.min(10, (frames - (i * 10)));
		var subframes = frames;
		var msTotal = 0;
		
		var pattern = {
			'frames': []
		};
		
		for (var f = 0; f < subframes; f++) {
			var frame = {
				'ms': msTotal,
				'zones': {}
			}
			var r = Math.ceil(Math.random() * 255);
			var g = Math.ceil(Math.random() * 255);
			var b = Math.ceil(Math.random() * 255);
			for (var z = 1; z < zones+1; z++) {
				frame.zones[z] = r +','+ g +','+ b;
			}
			pattern.frames.push(frame);
			//msTotal += 100 + Math.ceil(Math.random() * 5) * 100;
			msTotal += 1500;
		}
		
		//	Always end with an empty frame, otherwise the last not empty frame will not be shown (since it has zero duration)
		if (i == subpatterns - 1) {
			var frame = {
				'ms': msTotal,
				'zones': {}
			}
			for (var z = 1; z < zones+1; z++) {
				frame.zones[z] = 0 +','+ 0 +','+ 0;
			}
			pattern.frames.push(frame);
		}
	}
	return pattern;
}